/***************************************************************************
 *   Copyright (C) 2006 by Yoshihiro Ota
 *   ota@j.email.ne.jp
 *                                                                         
 *   This program is free software; you can redistribute it and/or modify  
 *   it under the terms of the GNU General Public License as published by  
 *   the Free Software Foundation; either version 2 of the License, or     
 *   (at your option) any later version.                                   
 *                                                                         
 *   This program is distributed in the hope that it will be useful,       
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
 *   GNU General Public License for more details.                          
 *                                                                         
 *   You should have received a copy of the GNU General Public License     
 *   along with this program; if not, write to the                         
 *   Free Software Foundation, Inc.,                                       
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             
 *
 ***************************************************************************
 *
 * Algorithm:
 *  Searching for files with the same contents is 3 stage process.
 *  Gather All File Information
 *  1. Traverse all files and directories given to the command line option.
 *  1.1 Obtain file information such as timestamps.
 *  1.2 Grope files by a pair of device id and file size.  Unless files are
 *      in the same grope, the content of files won't match.
 *  Match same content candidates
 *  2. Find files with same contents and link them together.
 *  2.1 First, sort a group by timestamp and so on.
 *  2.1.1 Timestamp is used as the first key so that the oldest file is kept.
 *  2.2 Check a pair of two files at a time in a group.
 *  2.2.1 Check if file status allows to combine files together.
 *  2.2.1.1 Check if owner informations are the same.
 *  2.2.1.2 Check if permissions are the same. // not done yet
 *  2.3.1.3 Check if their i-nodes are not same.
 *  2.3.1.4 Check if file hasn't modified since the program started.
 *  2.3 Read files by i-node and check if contents are the same.
 *      Files are only read upon necessity.
 *  2.3.1 Check first 512 bytes if they are the same.
 *  2.3.2 Calculate checksums and compare. 
 *  2.3.2.1 MD5 is required.
 *  2.3.2.2 SHAs are available in FreeBSD. // need to be configurable
 *  2.3.3 Unlink and link the newer one to the older one.
 *         Note that there may be more than one newer files but all newer
 *         files have the same i-node.  Since files are read by i-node but
 *         not filename, the same i-node file is read at most one.
 *  2.4 Cleanup memory allocated for a group.
 *  3 Restore timestamps.
 *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h> /* vfprintf */
#include <string.h>
#include <errno.h> /* errno */
#include <time.h> /* difftime */
#include <math.h> /* ceil */

#include <sys/param.h> /* MAXPATHLEN */

#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h> /* link/unlink */ /* *dir */

#include <utime.h> /* utime */
#include <dirent.h> /* dirent */

#include <signal.h> /* signal */

typedef enum { false, true } boolean;

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

// #define __DEBUG__
#if defined(__DEBUG__)
#define aprintf(txt, ...) printf("%s:" txt, __func__, __VA_ARGS__)
#define X(x)	fprintf(stderr, #x " %d\n", x)
#else
#define aprintf(txt, ...)
#define X(x)
#endif

#include <md5.h>
#ifdef HAVE_SHA
#include <sha.h>
#include <sha256.h>
#endif

#if LINK_MAX <= 0
#error LINK_MAX not defined
#endif
int link_max = LINK_MAX / 4; /* reserve some extra space for others */

typedef enum
{
	shell_i = 's', /* shell script */
	progress_i = 'P', /* stage progress */
	stats_i = 'S', /* status at end */
	action_i = 'A', /* unlink and link */
	domain_i = 'D', /* domain related */

	files_i = 'f', /* file processing */
	dirs_i = 'd', /* directory processing */
} logging_x;

typedef struct path_x
{
	char *name;
	boolean is_dir:1;
	boolean is_link_src:1;
	struct utimbuf time;
	struct path_x *next; /* stack: first-in, last-out */
} path_x;

typedef struct inode_x
{
	ino_t st_ino;
	off_t st_size;
	nlink_t st_nlink;

	char head[512];
	int head_len;

	boolean sig_ready; /* signature has been genereated */
	char tail[512];
	int tail_len;
#ifdef PRINT
	char md5sum[33];
  #ifdef HAVE_SHA
	char sha1sum[41];
	char sha256sum[65];
  #endif
#else
	unsigned char md5sum[16];
  #ifdef HAVE_SHA
	unsigned char sha1sum[20];
	unsigned char sha256sum[32];
  #endif
#endif
	struct inode_x *next;
} inode_x;

typedef struct
{
	path_x *path;
	struct stat stat;
	boolean stat_checked;
	inode_x *inode;
} file_x;

typedef struct
{
	dev_t st_dev; /* device ID */
	off_t st_size; /* file size */

	int num; /* for files */
	int max; /* for files */
	file_x **files; /* array of pointers */
	inode_x *inode; /* linkted-list stack */
} domain_x; /* device data */


typedef struct
{
	unsigned int execute:1;
	unsigned int restore_timestamp:1;
	unsigned int busy_fs:1;

	char *debug;
	int num; /* for domain */
	int max; /* for domain */
	domain_x **domain;
	path_x *access_time; /* stack of files and dirs : it maintains a list
		of directories and files to restore timestamps */
	struct
	{
		uint64_t files;
		uint64_t dirs;
		uint64_t linked;
	} stat;
} system_x;
system_x sys = { 1, 0, 0, NULL,
	0, 512, NULL, NULL, {0, }, };

void info(logging_x level, const char *format, ...)
{
	if(sys.debug != NULL && strchr(sys.debug, level))
	{
		va_list args;
		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
	}
}

domain_x* new_domain(dev_t st_dev, off_t st_size)
{
	domain_x *domain;
	domain = calloc(1, sizeof(*domain));
	domain->st_dev = st_dev;
	domain->st_size = st_size;
	domain->num = 0;
	domain->max = 16;
	domain->files = realloc(domain->files,
		domain->max * sizeof(*domain->files));
	info(domain_i, "new domain(%hu, %lld)\n", st_dev, st_size);
	return domain;
}

domain_x* find_domain(file_x *file)
{
	int i;
	for(i = 0; i < sys.num; i++)
		if(sys.domain[i]->st_dev == file->stat.st_dev &&
			sys.domain[i]->st_size == file->stat.st_size)
			return sys.domain[i];
	if(sys.num >= sys.max)
	{
		sys.domain =
			realloc(sys.domain,
				2 * sys.max * sizeof(sys.domain));
		sys.max = 2 * sys.max;
	}
	sys.domain[i] = new_domain(file->stat.st_dev, file->stat.st_size);
	sys.num++;
	return sys.domain[i];
}

inode_x* find_inode(domain_x *domain, ino_t st_ino)
{
	inode_x *inode = domain->inode;
	while(inode != NULL)
		if(st_ino == inode->st_ino)
			return inode;
		else
			inode = inode->next;
	return NULL;
}

int add_new_path(path_x *access_time)
{
	access_time->next = sys.access_time;
	sys.access_time = access_time;
	return 1;
}

void restore_timestamps(int sig)
{
	path_x *path;
	if(sig)
		fprintf(stderr, "signal %d\n", sig);

	while(path = sys.access_time)
	{
		if(path->is_dir || /* restore directory timestamps */
			path->is_link_src) /* restore file timestamps */
		{
			if(sys.restore_timestamp)
				utime(path->name, &path->time);
		}
		if(path->name)
			free(path->name);
		sys.access_time = path->next;
		free(path);
	}
	free(sys.domain);
	exit(sig);
}

int add_new_file(file_x *file)
{
	int i;
	domain_x * domain;
	domain = find_domain(file);

	/* XXX: check if NULL */
	if(domain->num >= domain->max)
	{
		domain->files =
			realloc(domain->files,
				2 * domain->max * sizeof(domain->files));
		domain->max = 2 * domain->max;
	}
	domain->files[domain->num++] = file;
	return 1;
}

int free_file(file_x **file)
{
	free((*file)->path->name);
	(*file)->path->name = NULL;
	free(*file);
	*file = NULL;
}

int compar(const void *data1, const void *data2)
{
	const file_x * file1 = *(file_x * const*) data1;
	const file_x * file2 = *(file_x * const*) data2;

#if 0
	if(file1->stat.st_size < file2->stat.st_size)
		return -1;
	else if(file1->stat.st_size > file2->stat.st_size)
		return 1;
/* compare uid and gid here so that sort does more work */
#endif
	return ceil(difftime(file1->stat.st_mtime, file2->stat.st_mtime));
}

int sort(domain_x *domain)
{
	qsort(domain->files, domain->num, sizeof(domain->files[0]), compar);
}

int file_not_linked(file_x *one, file_x *another)
{ /* if file is already processed, it will be NULL */
	if(one == NULL || another == NULL)
		return 0; /* false */
	return 1; /* true */
}

int file_stat_allowed(file_x *one, file_x *another)
{
	if(one->stat.st_dev != another->stat.st_dev)
		return 0; /* false */
	if(one->stat.st_uid != another->stat.st_uid)
		return 0; /* false */
	if(one->stat.st_gid != another->stat.st_gid)
		return 0; /* false */
	return 1; /* true */
}

/* check if the file is modified since the program is started.  If so, this
file is activly being modified and ignored from operation. */
boolean file_not_modified(file_x *file)
{
	struct stat stat;
	if(file->stat_checked)
		return true;
	if(lstat(file->path->name, &stat) == -1)
	{
		perror("file_not_modified:lstat");
		printf("path %s\n", file->path);
		return false;
	}
	if(ceil(difftime(file->stat.st_mtime, stat.st_mtime)) == 0)
	{
		file->stat_checked = true;
		return true;
	}
	return false;
}

int files_not_modified(file_x *one, file_x *another)
{
	if(!file_not_modified(one))
	{
		perror("file is modified while operation");
		return 0; /* false */
	}
	if(!file_not_modified(another))
	{
		perror("file is modified while operation");
		return 0; /* false */
	}
	return 1; /* true */
}

int read_head(domain_x *domain, file_x *file)
{
	FILE *fd;
	if(file->inode == NULL)
		file->inode = find_inode(domain, file->stat.st_ino);
	if(file->inode) /* already inode and head is read */
		return 1; /* true */
	if(!(file->inode = calloc(1, sizeof(*(file->inode)))))
	{
		perror("get_content:calloc");
		return 0; /* false */
	}
	file->inode->st_ino = file->stat.st_ino;
	file->inode->st_size = file->stat.st_size;
	file->inode->st_nlink = file->stat.st_nlink; /* this is the first time
		to refrence this file with this i-node. thus, it is safe
		to assume the number of hard links hasn't changed by this
		program. This is not a good assumption if other programs
		are running... */
	fd = fopen(file->path->name, "r");
	if(fd == NULL)
	{
		perror("get_content:fopen");
		return 0; /* false */
	}
	file->inode->head_len =
		fread(file->inode->head, 1, sizeof(file->inode->head), fd);
	fclose(fd);
	file->inode->next = domain->inode;
	domain->inode = file->inode;
	return file->inode->head_len; /* true */
}

int same_head(domain_x *domain, file_x *one, file_x *another)
{
	if(!read_head(domain, one))
		return 0;
	if(!read_head(domain, another))
		return 0;
	if(one->inode->head_len != another->inode->head_len)
		return 0;
	return !memcmp(one->inode->head, another->inode->head,
		one->inode->head_len);
}

int read_sig(file_x *file)
{
	FILE *fd;
	MD5_CTX md5;
#ifdef HAVE_SHA
	SHA1_CTX sha1;
	SHA256_CTX sha256;
#endif
	int len = 0;
	if(file->inode->sig_ready)
		return 1; /* true */
	/* inode->head contains all contents; no need to compare signatures */
	if(file->inode->st_size <= sizeof(file->inode->tail))
		return 1; /* true */
	fd = fopen(file->path->name, "r");
	if(fd == NULL)
	{
		perror("read_sig:fopen");
		return 0; /* false */
	}
	MD5Init(&md5);
#ifdef HAVE_SHA
	SHA1_Init(&sha1);
	SHA256_Init(&sha256);
#endif
	do
	{
		file->inode->tail_len = len;
		len = fread(file->inode->tail, 1, sizeof(file->inode->tail), fd);
		MD5Update(&md5, file->inode->tail, len);
#ifdef HAVE_SHA
		SHA1_Update(&sha1, file->inode->tail, len);
		SHA256_Update(&sha256, file->inode->tail, len);
#endif
	}while(!feof(fd));
	if(len != 0)
		file->inode->tail_len = len;
#ifdef PRINT
	MD5End(&md5, file->inode->md5sum);
#ifdef HAVE_SHA
	SHA1_End(&sha1, file->inode->sha1sum);
	SHA256_End(&sha256, file->inode->sha256sum);
#endif
#else
	MD5Final(file->inode->md5sum, &md5);
#ifdef HAVE_SHA
	SHA1_Final(file->inode->sha1sum, &sha1);
	SHA256_Final(file->inode->sha256sum, &sha256);
#endif
#endif
	fclose(fd);
#if 0
	aprintf("path(%s) -> md5(%s)\n", file->path, file->inode->md5sum);
#endif
	file->inode->sig_ready = true;
	return 1; /* true */
}

int same_contents(file_x *one, file_x *another)
{
	/* inode->head contains all contents; no need to compare signatures */
	if(one->inode->st_size <= sizeof(one->inode->tail))
		return 1; /* true */
	read_sig(one);
	read_sig(another);
	if(one->inode->tail_len != another->inode->tail_len)
		return 0; /* false */
	if(strcmp(one->inode->tail, another->inode->tail) != 0)
		return 0; /* false */
	if(memcmp(one->inode->md5sum, another->inode->md5sum,
		sizeof(one->inode->md5sum)) != 0)
		return 0; /* false */
#ifdef HAVE_SHA
	if(memcmp(one->inode->sha1sum, another->inode->sha1sum,
		sizeof(one->inode->sha1sum)) != 0)
		return 0; /* false */
	if(memcmp(one->inode->sha256sum, another->inode->sha256sum,
		sizeof(one->inode->sha256sum)) != 0)
		return 0; /* false */
#endif
	return 1; /* true */
}

int same_file(domain_x *domain, int one, int another)
{
	if(!file_not_linked(domain->files[one], domain->files[another]))
		return 0; /* false */

	if(domain->files[one]->stat.st_ino ==
		domain->files[another]->stat.st_ino) /* ignore same inode */
	{
		free_file(&domain->files[another]);
		return 0; /* it is same file but already has same inode */
	}

	if(!file_stat_allowed(domain->files[one], domain->files[another]))
		return 0; /* false */
	if(sys.busy_fs)
		if(!files_not_modified(domain->files[one],
			domain->files[another]))
			return 0; /* false */
	if(!same_head(domain, domain->files[one], domain->files[another]))
		return 0; /* false */
	if(!same_contents(domain->files[one], domain->files[another]))
		return 0; /* false */
	return 1;
}

int link_file(domain_x *domain, int one, int another)
{
	if(sys.execute)
	{
		if(domain->files[one]->inode->st_nlink >= link_max)
		{
			info(action_i, "link_max exceesed (%s[%u] %d)\n",
				domain->files[one]->path->name,
				domain->files[one]->stat.st_ino,
				domain->files[one]->stat.st_nlink);
			free_file(&domain->files[one]); /* check this later */
		}
		else
		{
			unlink(domain->files[another]->path->name);
			domain->files[one]->path->is_link_src = 1;
			link(domain->files[one]->path->name,
				domain->files[another]->path->name);
			sys.stat.linked++;
			info(action_i, "unlink(%s[%u]); link(%s[%u], %s[%u])\n",
				domain->files[another]->path->name,
				domain->files[another]->stat.st_ino,
				domain->files[one]->path->name,
				domain->files[one]->stat.st_ino,
				domain->files[another]->path->name,
				domain->files[another]->stat.st_ino);
			info(shell_i, "rm %s; ln %s %s\n",
				domain->files[another]->path->name,
				domain->files[one]->path->name,
				domain->files[another]->path->name);
			domain->files[one]->inode->st_nlink++;
			domain->files[another]->inode->st_nlink--;
			free_file(&domain->files[another]);
		}
	}
}

void free_domain(domain_x *domain)
{
	int i;
	inode_x *inode;
	info(domain_i, "free domain(%hu, %lld)\n",
		domain->st_dev, domain->st_size);
	for(i = 0; i < domain->num; i++)
		if(domain->files[i] != NULL)
		{
#if 0
			if(sys.restore_timestamp)
				utime(domain->files[i]->path->name,
					&domain->files[i]->path->time);
			if(domain->files[i]->path->name)
			{
				free(domain->files[i]->path->name);
				domain->files[i]->path->name = NULL;
			}
#endif
			free_file(&domain->files[i]);
		}
	while(domain->inode)
	{
		inode = domain->inode;
		domain->inode = domain->inode->next;
		free(inode);
	}
	free(domain);
}

int fision_by_domain(domain_x *domain)
{
	int i, j;
/* saved this for later enhancement.
	while(end < domain->num && domain->files[end]->stat.st_size == size)
		end++;
*/
	for(i = 0; i < domain->num; i++)
	{
		for(j = i+1;
			j < domain->num && domain->files[i] != NULL; j++)
		{
			if(domain->files[j] != NULL)
			{
				if(same_file(domain, i, j))
					link_file(domain, i, j);
			}
		}
	}
	free_domain(domain);
	return i;
}

int fision()
{
	int i = 0;
	info(progress_i, "1st stage completed: gathered all filenames\n");
	for(i = 0; i < sys.num; i++)
	{
		domain_x *domain = sys.domain[i];
		file_x *file = domain->files[0];
		sort(domain);
#if 0
		aprintf("dev%u ino%u uid%u gid%u mtime%s size%lu path%s\n",
			file->stat.st_dev, file->stat.st_ino,
			file->stat.st_uid, file->stat.st_gid,
			ctime(&file->stat.st_mtime), file->stat.st_size,
			file->path
			);
#endif
		info(domain_i,
			"process domain(%hu, %lld) %d files %d/%d domains\n",
			domain->st_dev, domain->st_size, domain->num,
			i, sys.num);
		fision_by_domain(domain);
	}
	info(progress_i, "2nd stage completed: combined same files\n");
	restore_timestamps(0);
	info(progress_i, "3rd stage completed: restored timestamps\n");
	info(stats_i,
		"%llu files\n"
		"%llu dirs\n"
		"%llu linked\n"
		,
		sys.stat.files,
		sys.stat.dirs,
		sys.stat.linked,
		0);
}


int loopdir(char *path,
	int (*callback)(char *parent, struct dirent *entry, void *data),
	void *data)
{
	DIR *dir;
     	struct dirent *entry;
	int i = 0;

	dir = opendir(path);
	if(dir != NULL)
	{
		while((entry = readdir(dir)) != NULL)
		{
			if(entry->d_name[0] == '.')
			{
				if(entry->d_name[1] == 0)
					continue;
				else if(entry->d_name[1] == '.' &&
					entry->d_name[2] == 0)
					continue;
			}
			callback(path, entry, data);
		}
		closedir(dir);
	}
}

int retreave(char *parent, struct dirent *entry, void *data)
{
	char path[MAXPATHLEN];
	path_x *orig_time;
	struct stat stat;

	orig_time = calloc(1, sizeof(*orig_time));
	if(entry)
		snprintf(path, sizeof(path), "%s/%s", parent, entry->d_name);
	else
		snprintf(path, sizeof(path), "%s", parent);

	if(lstat(path, &stat) == -1)
	{
		perror("lstat");
		return 99;
	}
	orig_time->name = strdup(path);
	orig_time->time.actime = stat.st_atime;
	orig_time->time.modtime = stat.st_mtime;
	if(S_ISREG(stat.st_mode))
	{
		file_x *file;

		info(files_i, "file %s\n", orig_time->name);
		if(!(file = calloc(1, sizeof(*file))))
		{
			perror("main:calloc");
			return errno;
		}
		add_new_path(orig_time);

		file->path = orig_time;
		file->stat = stat;
		add_new_file(file);
		sys.stat.files++;
	}
	else if(S_ISDIR(stat.st_mode))
	{
		orig_time->is_dir = true;
		add_new_path(orig_time);
		info(dirs_i, "dir %s\n", orig_time->name);
		loopdir(orig_time->name, retreave, NULL);
		sys.stat.dirs++;
	}
	else
	{
		free(orig_time->name);
		free(orig_time);
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int i;
	char c;

	while((c = getopt(argc, argv, "d:abhm:nt")) != -1)
	{
		switch(c)
		{
			case 'b':
				sys.busy_fs = 1;
				break;
			case 'd':
				sys.debug = optarg;
				break;
			case 'm':
				link_max = atoi(optarg);
				break;
			case 'n':
				sys.execute = 0;
				break;
			case 't':
				sys.restore_timestamp = 1;
				break;
			case 'h':
	fprintf(stderr, "ffunsion <files>\n"
		"-d : debug\n"
			"\t's' shell script (broken...)\n"
			"\t'P' stage progress\n"
			"\t'S' status at end\n"
			"\t'A' unlink and link\n"
			"\t'D' domain related\n"
			"\t'f' file processing\n"
			"\t'd' directory processing\n"
		"-b : busy file system where files are being updated\n"
		"-m : link_max\n"
		"-n : no-execute; dry-run\n"
		"-t : restore timestamps\n"
		"-h : help\n"
		"");
				exit(0);
			case '?':
			default:
				break;
		}
	}

	argc -= optind;
	argv += optind;
	signal(SIGINT, restore_timestamps); /* ctrl-C */
	signal(SIGQUIT, restore_timestamps); /* ctrl-\ : ??? */
	signal(SIGTERM, restore_timestamps); /* kill.1 */

	sys.domain = realloc(sys.domain, sys.max * sizeof(*sys.domain));

	for(c = 0; c < argc; c++)
		retreave(argv[c], NULL, NULL);
	fision();
	return 0;
}
